# -*- coding: utf-8 -*-

"""
　水稲を指定した日に県下全域に移植したとした時の出穂日と登熟日(暦日)の分布予測図を作成
するプログラム。

　このプログラムは、水稲の発育をDVI/DVR法で取り扱います。DVI/DVR法では、作物の発育状態を[0,1]の
発育指数(DVI)で表現し、発育の進行をDVIの増加で表現します。DVIの1日当たりの増分を発育速度(DVR)
と定義し、これと気温など気象値を関係づけることで作物の発育をモデル化し予測します。
　移植から出穂に至る発育と、出穂から登熟に至る発育とでは、パラメータは異なるものの、ともに積算温度型
の発育速度関数により定式化されているので、これを独自関数として定義し、パラメータを入れ替えて使用し
ます。
 　なお、DVI/DVR法の利用事例を見ると、特定の日のDVI値の計算の仕方に若干の不統一がみられます。
このプログラムでは、定植日のDVIを0とし、定植後のある日のDVIは、定植日からその日の前日までのDVRの
積算であるとしています。

　登熟日の分布がDrpに保存されるほか、出穂日の分布についてもDshに保存されるので、必要に応じこれらを
図化することもできます。

　ところで、今回のような、発育を面的に求める場合には、高山など作物を栽培できないメッシュが計算対象
に存在しうることを想定してプログラミングする必要があります。DVI/DVR法においては、何日積算しても
DVIが1にならない場合を想定してプログラミングします。
　このプログラムでは、メッシュ毎に、移植日からその年12月30日までについてDVRの積算を繰り返します。
そして、計算の途中でDVIが1となった場合には計算を中断し、その時の日付を記録します。このようにすると、
日々のDVRが小さくDVIが1に達しない場合でも12月31日に相当する日の計算で繰り返し計算は終了し、エ
ラーは発生しません。
　同様にして、成熟日の推定は、先の繰り返し計算で出穂が予測された翌日からその年の12月31日までについ
て繰り返します。

	OHNO, Hiroyuki 2018.02.21
"""

import numpy as np
import matplotlib.pylab as plt
import AMD_Tools3 as AMD
from datetime import datetime as dt
import copy

#発育速度関数の定義部分
def DVR(Ta, Para=[10.0, 1000.0]): #関数はdef文で定義します。
    """
    積算温度型の発育速度関数。
    引数(必須)：
      Ta：日平均気温
    引数(必要に応じ設定)  
      Para:基準温度(これ以下の場合積算をしない)、ならびに、出穂到達積算温度を束ねたリスト
           デフォルト値として[10, 1000]が設定されている。
    戻り値：気温に対する発育速度値

    注意：パラメータにデフォルトと異なる値を使用するときは、関数定義を書き換えるのではなく、
　　　　    関数を利用するときにその値を与えるようにします。
    """
    if Para[0] < Ta:
        DVR = (Ta-Para[0])/Para[1]
    else:
        DVR = 0.0
    return DVR 

#プログラムのメイン部分
if __name__ == "__main__":
    #移植条件の設定
    year = '2017'
    Day_of_Transplanting = '05-10'
    DVIini = 0.1
    phd = [10.0, 1050.0]    #移植～出穂に使用するDVR関数の品種パラメータ(サンプルです！)
    prp = [ 5.0,  850.0]   #出穂～登熟に使用するDVR関数の品種パラメータ(サンプルです！)

    #メッシュデータの取得
    prefecture = 'pref_1500'	#新潟県
    lalodomain = [36.733, 38.558, 137.627, 139.898]   #県の領域は、利用者Wikiの地理情報についての詳細ページの一覧表から知ることができます。
    timedomain = [year+'-'+Day_of_Transplanting, year+'-12-31' ]      #移植日から年末までを取得。
    Msh, tim, lat, lon = AMD.GetMetData('TMP_mea', timedomain, lalodomain)	#気象データを取得
    Pref, lat, lon = AMD.GetGeoData(prefecture, lalodomain) # 県のID　これも一覧表に掲載されています。
    
    #計算対象とすべきメッシュの選別(海や湖沼の排除)
    valimesh = ~np.isnan(Msh[0,:,:]*Pref) #期間初日の気象データがnp.nanでなく、かつ、県に属しているメッシュを対象とする
    y,x = np.where( valimesh )       #計算対象とすべきメッシュの緯度経度の組
    nome = len(y)                    #計算対象とするメッシュの数

    #気象データのサイズを控えておく
    ntim = Msh.shape[0] #メッシュデータの日数
    nlat = Msh.shape[1]	#メッシュの行数
    nlon = Msh.shape[2]	#メッシュの列数
    #出穂日に関する配列を用意する
    thd = np.zeros((nlat, nlon)) #出穂の段階となる日(起算日からの日数)を格納する配列
    thd[~valimesh] = np.nan          #計算対象でないメッシュにnp.nanを入れる
    Dhd = np.ma.zeros((nlat, nlon), dtype='datetime64[D]') #日付(numpyのdatetime64オブジェクト)を格納するndarrayはこのようにして生成します。
    Dhd = np.ma.masked_where(~valimesh, Dhd)    #datetime64オブジェクトには無効値が定義されていないので「mask」という技法で不要な領域を透明化します。
    #登熟日に関する配列を用意する
    trp = np.zeros((nlat, nlon)) #登熟の段階となる日(起算日からの日数)を格納する配列
    trp[~valimesh] = np.nan          #計算対象でないメッシュにnp.nanを入れる
    Drp = np.ma.zeros((nlat, nlon), dtype='datetime64[D]') #日付(numpyのdatetime64オブジェクト)を格納するndarrayはこのようにして生成します。
    Drp = np.ma.masked_where(~valimesh, Drp)    #datetime64オブジェクトには無効値が定義されていないので「mask」という技法で不要な領域を透明化します。

    #発育の計算
    dtp = tim[0]    #移植日(日付オブジェクト)
    ttp = 0         #移植日(移植日を0とする日数)
    for i in range(nome):	#対照メッシュを順次処理
        Ta = Msh[:,y[i],x[i]]  #メッシュ[yi,xi]における、1年分の気温を取り出す
        #出穂日の計算
        DVI = DVIini #移植時の発育指数
        for t in range(ttp+1,ntim-1):	#日付のループ(移植日翌日から12月30日まで回す)
            DVI += DVR(Ta[t-1],phd)   #ｔ日目におけるDVI
            if DVI > 1.0 : #DVIが1を超えたら
                break
        thd[y[i],x[i]] = t         #日数を保存
        Dhd[y[i],x[i]] = tim[t]    #日付を保存
        ttt = t #以降の繰り返し計算のスタートラインとして使うので保存
        #引き続き登熟日の計算
        DVI = 0.0
        for t in range(ttt+1, ntim):	#日付のループ(出穂日翌日から12月31日まで回す)
            DVI += DVR(Ta[t-1],prp)   #ｔ日目におけるDVI
            if DVI > 1.0 : 
                break
        trp[y[i],x[i]] = t         #日数を書き換える
        Drp[y[i],x[i]] = tim[t]    #日付を書き換える

    #以下、日付の分布図を書くのに重宝します。必要に応じてコピペして使用してください。－－－－－－－－
    Date2D = Dhd   #分布図はD2Dという名の変数に入れてください。
    #lat = lat      緯度はlatという名の変数に入れてください。
    #lon = lon      経度はLonという名の変数に入れてください。
    tate = 6      #図を入れるオブジェクト(入れ物)の全体的な大きさを指定します。
    figtitle ="Date of Heading (for t/p on "+Day_of_Transplanting+" )"
    manualscl = True    #カラースケールの上限値と下限値を指定したいときに使用します。
    sclmax = np.datetime64("2017-10-31")		#日付の最大値
    sclmin = np.datetime64("2017-07-26")		#日付の最小値
    sclint = np.timedelta64(1,"D")        #日付の刻み
    #－－－－－－－－
    if not manualscl :
        sclmax = np.max(Date2D)
        sclmin = np.min(Date2D)
        sclint = (sclmax-sclmin)/20.0
    yoko = tate * (np.max(lon)-np.min(lon))/(np.max(lat)-np.min(lat)) + 2
    fig = plt.figure(num=None, figsize=(yoko, tate))#図を入れるオブジェクト(入れ物)を定義
    plt.axes(facecolor='0.8')    #背景を灰色に
    levels = np.arange(sclmin,sclmax+sclint, sclint)
    #cmap = plt.cm.get_cmap("Spectral").copy()    #色調(カラーマップ)を愛称(「_r」を最後に付けると反転する)で指定
    cmap = copy.copy(plt.cm.get_cmap("Spectral"))     #色調(カラーマップ)を愛称(「_r」を最後に付けると反転する)で指定
    cmap.set_over('w', 1.0)   #上限を超過色に白を指定
    cmap.set_under('k', 1.0)  #下限を超過色に黒を指定
    CF = plt.contourf(lon, lat, Date2D, levels, cmap=cmap, extend='both')
    CB = plt.colorbar(CF, format=plt.DateFormatter('%b %d'))				#カラーバーを描く
    plt.title(figtitle)
    plt.savefig('Heading'+'.png', dpi=600) 	#図の画像ファイル出力
    plt.show(fig)									#図の表示
    plt.clf()
    #分布図ここまで－－－－－－－－－－－－－－－－－－－－－－－－－－－－－－－

    #地理院地図表示用ファイル出力
    AMD.PutGSI_Map(Dhd,lat,lon,cmapstr="Spectral",minmax=[sclmin,sclmax],
                    label="Date in Heading",filename="Heading")
    
    #以下、日付の分布図を書くのに重宝します。必要に応じてコピペして使用してください。－－－－－－－－
    Date2D = Drp   #分布図はD2Dという名の変数に入れてください。
    #lat = lat      緯度はlatという名の変数に入れてください。
    #lon = lon      経度はLonという名の変数に入れてください。
    tate = 6      #図を入れるオブジェクト(入れ物)の全体的な大きさを指定します。
    figtitle ="Date in Mature (for t/p on "+Day_of_Transplanting+" )"
    manualscl = True    #カラースケールの上限値と下限値を指定したいときに使用します。
    sclmax = np.datetime64("2017-10-31")		#日付の最大値
    sclmin = np.datetime64("2017-07-26")		#日付の最小値
    sclint = np.timedelta64(1,"D")        #日付の刻み
    #－－－－－－－－
    if not manualscl :
        sclmax = np.max(Date2D)
        sclmin = np.min(Date2D)
        sclint = (sclmax-sclmin)/20.0
    yoko = tate * (np.max(lon)-np.min(lon))/(np.max(lat)-np.min(lat)) + 2
    fig = plt.figure(num=None, figsize=(yoko, tate))#図を入れるオブジェクト(入れ物)を定義
    plt.axes(facecolor='0.8')    #背景を灰色に
    levels = np.arange(sclmin,sclmax+sclint, sclint)
    #cmap = plt.cm.get_cmap("Spectral").copy()    #色調(カラーマップ)を愛称(「_r」を最後に付けると反転する)で指定
    cmap = copy.copy(plt.cm.get_cmap("Spectral"))     #色調(カラーマップ)を愛称(「_r」を最後に付けると反転する)で指定
    cmap.set_over('w', 1.0)   #上限を超過色に白を指定
    cmap.set_under('k', 1.0)  #下限を超過色に黒を指定
    CF = plt.contourf(lon, lat, Date2D, levels, cmap=cmap, extend='both')
    CB = plt.colorbar(CF, format=plt.DateFormatter('%b %d'))				#カラーバーを描く
    plt.title(figtitle)
    plt.savefig('Mature'+'.png', dpi=600) 	#図の画像ファイル出力
    plt.show(fig)									#図の表示
    plt.clf()
    #分布図ここまで－－－－－－－－－－－－－－－－－－－－－－－－－－－－－－－

    #地理院地図表示用ファイル出力
    AMD.PutGSI_Map(Drp,lat,lon,cmapstr="Spectral",minmax=[sclmin,sclmax],
                    label="Date in Mature",filename="Mature")